
----                                                 ----
-- addhook.lua                                         --
-- enable local and anonymous functions                --
-- for addhook, freehook, timer and freetimer          --
-- e.g. addhook("say", function(id, txt) msg(txt) end) --
----                                                 ----

-----------------------
--   GENERAL STUFF   --
-----------------------

-- Get the name of a function
local function GetFuncName(func)
	local n, t -- name & type
	
	-- Is it a global?
	for k, v in pairs(_G) do
		if v == func then n, t = k, "function " end
	end
	
	-- or is it a local?
	if not n then
		local i = 1 -- iterate over locals
		while true do
			local k, v = debug.getlocal(3, i)
			if not k then break end
			if v == func then n, t = k, "local function " break end
			i = i + 1
		end
	end
	
	-- Nope, it's an anonymous function
	if not n then local d = debug.getinfo(func, "S") n, t = d.short_src .. ":" .. d.linedefined, "anonymous function " end
	
	return n, t
end




-- Get function by name
local function GetFuncByName(func)
	local f, t -- function & type
	
	-- Check for global first
	if _G[func] and type(_G[func]) == "function" then
		f, t = _G[func], "function "
		
	-- Nope, it's no global :(, check for local
	else
		t = "local function "
		local i = 1 -- iterate over locals
		while true do
			local k, v = debug.getlocal(3, i) -- name, value
			if not k then break end
			if k == func then f, t = v, "local function " break end
			i = i + 1
		end
			
				-- Not a local either
		if not f or type(f) ~= "function" then
			error("GetFuncByName: function " .. func .. " could not be found.", 3)
			return
		end
	end
	
	return f, t
end





-----------------------
--   ADDHOOK STUFF   --
-----------------------

-- Create a local copy of the original addhook function,
-- so we can safely overwrite the function
local _addhook  = addhook
local _freehook = freehook


-- The table in which all hook functions are stored.
__ADDED_HOOKS__ = {} -- Better keep this sort of unique




-- Based on priority, call all functions hooked to this action
-- and return the return value of the function with the lowest
-- priority to CS2D.
local function callfunc(t, ...)
	local ret, m = {}, math.huge
	for i, v in pairs(t) do if i < m then m = i end ret[i] = v(...) end
	return ret[m] or 0
end



-- If no function has been hooked to this action before, create a hew hook
-- __ADDED_HOOKS__[k] is both a table and a function; if called, the local
-- function callfunc(line 16) will be called.
local function newhook(t, k)
	__ADDED_HOOKS__[k] = {}
	setmetatable(__ADDED_HOOKS__[k], { __call = callfunc })
	_addhook(k, "__ADDED_HOOKS__." .. k)
	return __ADDED_HOOKS__[k]
end


-- If a non-existing index is requested (by table.insert),
-- the local function newhook will create a new hook and return
-- an empty table (with metatable).
setmetatable(__ADDED_HOOKS__, { __index = newhook })




-- Rewrite the addhook function
-- Most of what you see here (all the debug stuff) is retrieving the function name,
-- just to be able to print "Lua: adding (local | anonymous) function 'name | file:line' to hook 'hook'"
function addhook(hook, func, priority)
	local l, f, n = "function " -- function 'type' (function, local function), the function itself and the function name (or description)

	-- Handle (anonymous) functions
	if type(func) == "function" then
		f, n, l = func, GetFuncName(func)
	
	-- Handle literal function names
	elseif type(func) == "string" then
		n, f, l = func, GetFuncByName(func)
	
	-- Some other type?
	else
		error("addhook: " .. tostring(func) .. " is not a valid function (type '" .. type(func) .. "')", 2)
		return
	end
	
	-- Add hook to hooks table
	if priority then table.insert(__ADDED_HOOKS__[hook], priority, f)
	else table.insert(__ADDED_HOOKS__[hook], f) end
	print("Lua: Adding " .. l .. "'" .. n .. "' to hook '" .. hook .. "'")
end


function freehook(hook, func)
	local l, f, n = "function "
	
	-- Handle (anonymous) functions
	if type(func) == "function" then
		f, n, l = func, GetFuncName(func)
	
	-- Blah blah you get the point
	elseif type(func) == "string" then
		n, f, l = func, GetFuncByName(func)
	
	else
		error("addhook: " .. tostring(func) .. " is not a valid function (type '" .. type(func) .. "')", 2)
		return
	end

	print("Lua: Removing " .. l .. "'" .. n .. "' from hook '" .. hook .. "'")	
	for i, v in pairs(__ADDED_HOOKS__[hook]) do
		if v == f then table.remove(__ADDED_HOOKS__[hook], i) break end
	end
	
	-- If that was the last function hooked to the action, remove the hook altogether
	if #__ADDED_HOOKS__[hook] == 0 then _freehook(hook, "__ADDED_HOOKS__." .. hook) __ADDED_HOOKS__[hook] = nil end
end








-----------------------
--    TIMER STUFF    --
-----------------------

local _timer = timer
local _freetimer = freetimer

__ADDED_TIMERS__ = {}

function timer(time, func, p, c)
	
	if type(func) == "function" then func = func
	elseif type(func) == "string" then func = GetFuncByName(func)
	else
		error("timer: " .. tostring(func) .. " is not a valid function (type " .. type(func))
		return
	end
	
	-- Quite unique key: function pointer + argument
	local k = (tostring(func) .. tostring(p)):gsub("[^a-zA-Z0-9_]*", "_")
	__ADDED_TIMERS__[k] = { func = func, p = p }
	
	_timer(time, "__ADDED_TIMERS__." .. k .. ".func", p, c)
end


function freetimer(func, p)

	if type(func) == "function" then func = func
	elseif type(func) == "string" then func = GetFuncByName(func)
	else
		error("timer: " .. tostring(func) .. " is not a valid function (type " .. type(func))
		return
	end
	
	for k, v in pairs(__ADDED_TIMERS__) do
		if v.func == func and v.p == p then
			_freetimer("__ADDED_TIMERS__." .. k .. ".func", v.p)
			break
		end
	end
end

-- Banaan @ 04-01-12
